# coding: utf-8

# fields names
ID = "id"
KEY = "key"
INDEX = "index"
FIGURES = "figures"
OBJECT_ID = "objectId"
OBJECT_KEY = "objectKey"
FRAME_RANGE = "frameRange"

TAGS_MAP = "tags"
OBJECTS_MAP = "objects"
FIGURES_MAP = "figures"
VIDEOS_MAP = "videos"

FRAMES = "frames"

OBJECTS = "objects"
FRAMES_COUNT = "framesCount"
VIDEO_ID = "videoId"
VIDEO_NAME = "videoName"
TAGS = "tags"
DESCRIPTION = "description"

IMG_DESCRIPTION = "description"
IMG_SIZE = "size"
IMG_SIZE_WIDTH = "width"
IMG_SIZE_HEIGHT = "height"

META = "meta"
